@@ -78,9 +78,6 @@ impl WebBackend {
 /// Data gathered between frames.
 #[derive(Default)]
 pub struct WebInput {
-    /// Is this a touch screen? If so, we ignore mouse events.
-    pub is_touch: bool,
-
     /// Required because we don't get a position on touched
     pub latest_touch_pos: Option<egui::Pos2>,
 
